"""Tools Utils Module"""
def filter_tools(keys, dictionary):
    """
    Fetches values from the dictionary based on provided keys.

    Args:
    - dictionary (dict): The source dictionary.
    - keys (list): List of keys to fetch values for.

    Returns:
    - list: List of values corresponding to the provided keys.
    """
    return [dictionary.get(key) for key in keys]

def format_agent_actions(steps: list[tuple]) -> list[dict]:
    return [
        {"tool": step[0].tool, "tool_input": step[0].tool_input, "log": step[0].log}
        for step in steps
    ]

def match_strings(keys: list[str], functions):
    """Match the strings in the keys to the functions"""
    # Initialize array to store output
    output = []

    # Loop through the functions array
    for function in functions:
        # If name property of function matches one of the strings in keys
        if function['name'] in keys:
            # Append the function to the output array
            output.append(function)

    # Return the output array
    return output


"""Request Models"""
from typing import Any, List, Optional, Union
from pydantic import BaseModel, Field

from promptengineers.models.loader import (
    TypeUrlLoader,
    YoutubeLoader,
    BlockchainLoader,
    CopyPasteLoader,
)

class RequestMultiLoader(BaseModel):
    index_name: str = Field(...)
    provider: str = ("pinecone", "redis")
    embedding: str = ("text-embedding-ada-002", "llama2:7b", "llama2")
    files: List[str] or None = Field(...)
    loaders: List[
        Union[TypeUrlLoader, YoutubeLoader, BlockchainLoader, CopyPasteLoader]
    ] or None = Field(...)

    class Config:
        json_schema_extra = {
            "example": {
                "provider": "pinecone",
                "embedding": "text-embedding-ada-002",
                "index_name": "formio-docs-and-website",
                "files": [
                    "formio-customer-issue.pdf",
                ],
                "loaders": [
                    {"type": "gitbook", "urls": ["https://help.form.io"]},
                    {"type": "web_base", "urls": ["https://form.io"]},
                ],
            }
        }


class RequestDataLoader(BaseModel):
    index_name: str = Field(...)
    provider: str = ("pinecone", "redis")
    embedding: str = ("text-embedding-ada-002", "llama2:7b", "llama2")
    loaders: List[
        Union[TypeUrlLoader, YoutubeLoader, BlockchainLoader, CopyPasteLoader]
    ] or None = Field(...)

    class Config:
        json_schema_extra = {
            "example": {
                "provider": "pinecone",
                "embedding": "text-embedding-ada-002",
                "index_name": "formio-docs-and-website",
                "loaders": [
                    {"type": "gitbook", "urls": ["https://help.form.io"]},
                    {"type": "web_base", "urls": ["https://form.io"]},
                    {"type": "copy", "text": "This is a test."},
                ],
            }
        }


CHAT_MESSAGES = [
	{"role": "system", "content": "You are a helpful assistant."},
	{"role": "user", "content": 'Who won the 2001 world series?'},
	{"role": "assistant", "content": 'The arizona diamondbacks won the 2001 world series.'},
	{"role": "user", "content": 'Who were the pitchers?'},
]

FUNCTION_MESSAGES = [
	{"role": "system", "content": "You are a helpful assistant."},
	{"role": "user", "content": 'What is the length of supercalifragilisticexpialidocious?'},
]



from typing import List

from promptengineers.models.prompt import PromptSystem

STRIPE_API_DOC_PROMPT = {
	"title": "Stripe Document Retreival Prompt",
	"system": "You are a helpful {ASSISTANT_TYPE} assistant. You have have been given context to {INDEX_NAME} to answer user questions.",
	"variables": {
		"ASSISTANT_TYPE": 'document retreival',
		"INDEX_NAME": 'Stripe API docs'
	}
}

class ResponsePromptSystemList(PromptSystem): # pylint: disable=too-few-public-methods
	prompts: List[PromptSystem] = []

	class Config: # pylint: disable=too-few-public-methods
		"""Prompt System Template"""
		json_schema_extra = {
			"example": {
				"prompts": [STRIPE_API_DOC_PROMPT]
			}
		}

class ResponsePromptSystem(PromptSystem): # pylint: disable=too-few-public-methods

	class Config: # pylint: disable=too-few-public-methods
		"""Prompt System Template"""
		json_schema_extra = {
			"example": {
				"prompt": STRIPE_API_DOC_PROMPT
			}
		}


from typing import Optional
from pydantic import BaseModel

class Retrieval(BaseModel):
    """Contains the information needed to document retrieval augmented generation."""

    provider: Optional[str] = None
    index: Optional[str] = None

    class Config:  # pylint: disable=too-few-public-methods
        """Contains the information needed to document retrieval augmented generation."""

        json_schema_extra = {
            "example": {
                "provider": "pinecone",
                "index": "Formio",
            }
        }


"""Utilites for Chains"""
from langchain_core.documents.base import Document

def get_chat_history(inputs: tuple) -> str:
    """Formats the chat history into a readable format for the chatbot"""
    res = []
    for human, assistant in inputs:
        res.append(f"Human: {human}\nAssistant: {assistant}")
    return "\n".join(res)

def combine_documents(documents: list[Document]):
    combined_content = ""
    for doc in documents:
        combined_content += doc.page_content + "\n"
    return combined_content

def retrieve_system_message(messages):
    """Retrieve the system message"""
    try:
        filtered_messages = list(filter(lambda message: message['role'] == 'system', messages))
        return filtered_messages[0]['content']
    except IndexError:
        return None

def retrieve_chat_messages(messages):
    """Retrieve the chat messages"""
    return [
        (msg["content"]) for msg in messages if msg["role"] in ["user", "assistant"]
    ]


from .history import HistoryController
from .auth import AuthController
from .storage import StorageController
from .chat import ChatController
from .retrieval import VectorSearchController
from .prompt import PromptController
from .settings import SettingsController

__all__ = [
	'HistoryController',
	'AuthController',
	'StorageController',
	'ChatController',
	'VectorSearchController',
    'PromptController',
    'SettingsController'
]


from promptengineers.models.history import HistoryDocument, HISTORY_DOCUMENT

class ResponseHistoryIndex(HistoryDocument): # pylint: disable=too-few-public-methods
	histories: list[HistoryDocument]

	class Config: # pylint: disable=too-few-public-methods
		"""Response body for history index"""
		json_schema_extra = {
			"example": {
				"histories": [HISTORY_DOCUMENT]
            }
		}

class ResponseHistoryShow(HistoryDocument): # pylint: disable=too-few-public-methods
	history: HistoryDocument

	class Config: # pylint: disable=too-few-public-methods
		"""Request body for history show"""
		json_schema_extra = {
			"example": {
				"history": HISTORY_DOCUMENT
            }
		}


from .chat import router as chat_router
from .history import router as history_router
from .storage import router as storage_router
from .retrieval import router as retrieval_router
from .prompt import router as prompt_router
from .settings import router as settings_router

__all__ = [
    'chat_router',
    'history_router',
    'storage_router',
    'retrieval_router',
    'prompt_router',
    'settings_router'
]


from abc import ABC, abstractmethod

from langchain.llms import Ollama
from langchain.chat_models import ChatOpenAI

# Define the strategy interface
class LLMStrategy(ABC):
	@abstractmethod
	def chat(
		self,
		model_name: str,
		temperature: float or int = 0.9,
		streaming: bool = False,
		callbacks: list or None = None,
	):
		pass

class OpenAIStrategy(LLMStrategy):
	def __init__(self, api_key):
		self.api_key = api_key or ''

	def chat(
		self,
		model_name: str,
		temperature: float or int = 0.9,
		streaming: bool = False,
		callbacks: list or None = None,
	) -> str:
		return ChatOpenAI(
			openai_api_key=self.api_key,
			model_name=model_name,
			temperature=temperature,
			streaming=streaming,
			callbacks=callbacks,
		)

class BedrockStrategy(LLMStrategy):
	def __init__(self, s3_access_key, s3_secret_key):
		import boto3
		self.client = boto3.client(
			service_name='bedrock',
			aws_access_key_id=s3_access_key,
			aws_secret_access_key=s3_secret_key,
		)
	def chat(
		self,
		model_name: str,
		temperature: float or int = 0.9,
		streaming: bool = False,
		callbacks: list or None = None,
	) -> str:
		from langchain.chat_models import BedrockChat
		# Similarly, this would contain logic specific to another chat model provider
		return BedrockChat(
			client=self.client,
			region_name="us-east-1",
			model_id=model_name,
			model_kwargs={
				"temperature": temperature,
			},
			streaming=streaming,
			callbacks=callbacks,
		)

class OllamaStrategy(LLMStrategy):
	def __init__(self, base_url: str = 'http://localhost:11434'):
		self.base_url = base_url

	def chat(
		self,
		model_name: str,
		temperature: float or int = 0.9,
		streaming: bool = False,
		callbacks: list or None = None,
	) -> str:
		return Ollama(
			base_url=self.base_url,
			model=model_name,
			temperature=temperature,
			callbacks=callbacks,
		)

class ModelContext:
	def __init__(self, strategy: LLMStrategy):
		self.strategy = strategy

	def chat(
		self,
		model_name: str,
		temperature: float = None,
		streaming: bool = False,
		callbacks: list or None = None,
	) -> str:
		return self.strategy.chat(
			model_name=model_name,
			temperature=temperature,
			streaming=streaming,
			callbacks=callbacks,
		)



from typing import Optional
from pydantic import BaseModel

from promptengineers.models import Retrieval

CHAT_SETTINGS = {
	"title": "Prompt Engineers Core Repository Agent Settings",
    "model": "gpt-3.5-turbo",
    "system": """You are a helpful assistant. You are equipped with the following tools: {tools} for accomplishing tasks. If you to ask questions to perform a task, you can ask {vectorstore} for assistance.""",
    "temperature": 0.5,
	"stream": True,
    "tools": ["calculator", "github_new_issue", "github_edit_issue", "github_create_pull_request"],
    "plugins": ["https://domain-to-plugin-api.com/.well-known/ai-plugin.json"],
    "retrieval": {
		"provider": "pinecone",
		"index": "promptengineers-core-repository",
    },
	"tags": ["core", "repository", "promptengineers"]
}

CHAT_SETTING_DOCUMENT = {
	"_id": "532e147a126c8e67d951f234",
	"title": "Prompt Engineers Core Repository Agent Settings",
    "model": "gpt-3.5-turbo",
    "system": """You are a helpful assistant. You are equipped with the following tools: {tools} for accomplishing tasks. If you to ask questions to perform a task, you can ask {vectorstore} for assistance.""",
    "temperature": 0.5,
	"stream": True,
    "tools": ["calculator", "github_new_issue", "github_edit_issue", "github_create_pull_request"],
    "plugins": ["https://domain-to-plugin-api.com/.well-known/ai-plugin.json"],
    "retrieval": {
		"provider": "pinecone",
		"index": "promptengineers-core-repository",
    },
	"tags": ["core", "repository", "promptengineers"],
	"created_at": 1698523723,
	"updated_at": 1698562747
}

class ChatSettings(BaseModel): # pylint: disable=too-few-public-methods
	"""Chat Settings Template"""
	title: str
	model: str
	system: str
	temperature: Optional[float] = 0.5
	stream: Optional[bool] = False
	tools: Optional[list[str]] = None
	plugins: Optional[list[str]] = None
	retrieval: Retrieval = None
	tags: Optional[list[str]] = None

	class Config: # pylint: disable=too-few-public-methods
		"""Agent Settings Template"""
		json_schema_extra = {
			"example": CHAT_SETTINGS
		}

class ChatSettingDocument(ChatSettings): # pylint: disable=too-few-public-methods
	"""Chat Settings Template"""
	"""A message to send to the chatbot."""
	_id: str
	created_at: int
	updated_at: int

	class Config: # pylint: disable=too-few-public-methods
		"""Agent Settings Template"""
		json_schema_extra = {
			"example": CHAT_SETTING_DOCUMENT
		}

class ChatSettingIndex(ChatSettingDocument): # pylint: disable=too-few-public-methods
	"""Chat Settings Template"""
	"""A message to send to the chatbot."""
	settings: list[ChatSettings]

	class Config: # pylint: disable=too-few-public-methods
		"""Agent Settings Template"""
		json_schema_extra = {
			"example": {
				"settings": [CHAT_SETTING_DOCUMENT]
			}
		}


import requests
from urllib.parse import urljoin, urlparse

from bs4 import BeautifulSoup
from langchain.callbacks import get_openai_callback
from langchain.embeddings import OpenAIEmbeddings
from langchain.text_splitter import CharacterTextSplitter, RecursiveCharacterTextSplitter
from langchain.vectorstores import FAISS

from promptengineers.core.utils import logger

def get_links(url: str):
	response = requests.get(url, timeout=5)
	soup = BeautifulSoup(response.text, 'html.parser')
	links = []
	for link in soup.find_all('a'):
		href = link.get('href')
		if href and urlparse(href).netloc == '':
			links.append(urljoin(url, href))
	return links

def split_docs(
    documents,
    chunk_size: int = 1000,
    chunk_overlap: int = 100,
    recursive: bool = True,
):
    ## Text Spliter
    if recursive:
        text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap
        )
    else:
        text_splitter = CharacterTextSplitter(
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap
        )
    docs = text_splitter.split_documents(documents)
    return docs

def create_faiss_vectorstore(
    docs,
    chunk_size: int = 1000,
    chunk_overlap: int = 0
):
    """Load the vectorstore."""
    with get_openai_callback() as callback:
        logger.info('[utils.retrieval.create_vectorstore] Tokens: %s', callback.total_tokens)
        embeddings = OpenAIEmbeddings(max_retries=2)
        return FAISS.from_documents(
            split_docs(docs, chunk_size, chunk_overlap),
            embeddings
        )



"""Request Models"""
from typing import Any, List, Optional, Union
from pydantic import BaseModel, Field

from promptengineers.models import Retrieval

#################################################
## ChatGPT
#################################################
class ReqBodyChat(BaseModel):  # pylint: disable=too-few-public-methods
    """A message to send to the chatbot."""

    title: Optional[str] = None
    model: Optional[str] = None
    messages: Optional[Any] = None
    temperature: Optional[float or int] = None
    stream: Optional[bool] = None

    class Config:  # pylint: disable=too-few-public-methods
        """A message to send to the chatbot."""

        schema_extra = {
            "example": {
                "model": "gpt-3.5-turbo",
                "temperature": 0.8,
                "stream": False,
                "messages": [
                    {"role": "system", "content": "You are a helpful assistant."},
                    {"role": "user", "content": "Who won the 2001 world series?"},
                    {
                        "role": "assistant",
                        "content": "The arizona diamondbacks won the 2001 world series.",
                    },
                    {"role": "user", "content": "Who were the pitchers?"},
                ],
            }
        }

class ReqBodyAgentChat(ReqBodyChat):  # pylint: disable=too-few-public-methods
    """A message to send to the chatbot."""

    tools: list[str] = None
    plugins: list[str] = None
    retrieval: Retrieval = None

    class Config:  # pylint: disable=too-few-public-methods
        """A message to send to the chatbot."""

        schema_extra = {
            "example": {
                "model": "gpt-3.5-turbo",
                "temperature": 0.8,
                "stream": False,
                "messages": [
                    {"role": "system", "content": "You are a helpful assistant."},
                    {
                        "role": "user",
                        "content": "What will $2000 compounded at 5 percent for 10 years be?",
                    },
                ],
                "tools": ["math_tool"],
                "plugins": ["https://api.speak.com/.well-known/ai-plugin.json"],
                "retrieval": {
                    "provider": "pinecone",
                    "index": "Formio",
                }
            }
        }


class ReqBodyAgentPluginsChat(ReqBodyChat):  # pylint: disable=too-few-public-methods
    """A message to send to the chatbot."""

    plugins: list[str] = None

    class Config:  # pylint: disable=too-few-public-methods
        """A message to send to the chatbot."""

        schema_extra = {
            "example": {
                "model": "gpt-3.5-turbo-16k",
                "temperature": 0.8,
                "stream": False,
                "messages": [
                    {"role": "system", "content": "You are a helpful assistant."},
                    {
                        "role": "user",
                        "content": "Using the Speak tool, how should I politely greet shop employees when I enter, in French?",
                    },
                ],
                "plugins": ["https://api.speak.com/.well-known/ai-plugin.json"],
            }
        }

class ReqBodyVectorstoreChat(ReqBodyChat):  # pylint: disable=too-few-public-methods
    """A message to send to the chatbot."""

    vectorstore: Optional[str] = None
    provider: Optional[str] = None

    class Config:  # pylint: disable=too-few-public-methods
        """A message to send to the chatbot."""

        schema_extra = {
            "example": {
                "provider": "pinecone",
                "vectorstore": "Formio",
                "model": "gpt-3.5-turbo",
                "temperature": 0.8,
                "stream": False,
                "messages": [
                    {"role": "system", "content": "You are a helpful assistant."},
                    {"role": "user", "content": "Can you summarize the context?"},
                ],
            }
        }


class ReqBodyFunctionChat(ReqBodyChat):  # pylint: disable=too-few-public-methods
    """A message to send to the chatbot."""

    functions: list[str] = []

    class Config:  # pylint: disable=too-few-public-methods
        """A message to send to the chatbot."""

        schema_extra = {
            "example": {
                "model": "gpt-3.5-turbo",
                "temperature": 0.8,
                "messages": [
					{"role": "system", "content": "You are a helpful assistant."},
					{"role": "user", "content": 'What is the length of supercalifragilisticexpialidocious?'},
				],
                "functions": ["get_word_length"],
            }
        }


"""Retrieves model from OpenAI API and returns a ChatOpenAI object."""
import os

import openai

from promptengineers.core.config import OPENAI_API_KEY
from promptengineers.core.config.functions import FUNCTIONS
from promptengineers.functions.factories import FunctionFactory
from promptengineers.tools.utils import match_strings

os.environ['OPENAI_API_KEY'] = OPENAI_API_KEY

def openai_chat_model(
    model_name: str,
    streaming: bool = False,
    temperature: float or int = 0,
    messages: list or None = None,
    functions: list or None = None,
    function_call: str or None = None,
):
    """Query OpenAI API for a chat model."""
    if functions:
        return openai.ChatCompletion.create(
			messages=messages,
			model=model_name,
			temperature=temperature,
			functions=functions,
			function_call=function_call,
		)
    return openai.ChatCompletion.create(
		messages=messages,
		model=model_name,
		temperature=temperature,
		stream=streaming,
    )

def openai_chat_functions_model(
	model_name: str,
    streaming: bool = False,
    temperature: float or int = 0,
	messages: list or None = None,
    keys: list or None = None,
):
    """Query OpenAI API for a chat model."""
    call_fn = openai_chat_model(
		model_name=model_name,
        messages=messages,
        functions=match_strings(keys, FUNCTIONS),
        function_call="auto",
	)
    response_message = call_fn["choices"][0]["message"]
    if response_message.get("function_call"):
        function_name = response_message["function_call"]["name"]
        function_response = FunctionFactory().get_result(
			function_name,
			response_message,
		)
        # Step 4: send the info on the function call and function response to GPT
        messages.append(response_message)  # extend conversation with assistant's reply
        messages.append({
			"role": "function",
			"name": function_name,
			"content": function_response,
		})
    return openai_chat_model(
		messages=messages,
		model_name=model_name,
		temperature=temperature,
		streaming=streaming,
	)


"""Response models for the API.""" ""
from pydantic import BaseModel, Field


class ResponseStatus(BaseModel):
    """A message to send to the chatbot."""
    version: str = Field(default="v0.0.15")


#################################################
## Chat HTTP Response
#################################################
class ResponseChat(BaseModel):
    """A message to send to the chatbot."""

    class Config:  # pylint: disable=too-few-public-methods
        """A message to send to the chatbot."""

        json_schema_extra = {
            "example": {
				"message": "The pitchers for the Arizona Diamondbacks in the 2001 World Series were Randy Johnson, Curt Schilling, Miguel Batista, and Brian Anderson.",
				"usage": {
					"total_tokens": 80,
					"prompt_tokens": 50,
					"completion_tokens": 30,
					"total_cost": 0.000135,
					"successful_requests": 1
				}
			}
        }


#################################################
## Chat Agent HTTP Response
#################################################
class ResponseAgentChat(BaseModel):
    class Config:
        json_schema_extra = {
            "example": {
                "message": "$2000 compounded at 5 percent for 10 years will be approximately $3257.79.",
                "usage": {
                    "total_tokens": 245,
                    "prompt_tokens": 193,
                    "completion_tokens": 52,
                    "total_cost": 0.00039349999999999997,
                    "successful_requests": 2,
                },
            }
        }


#################################################
## Chat Agent Plugins HTTP Response
#################################################
class ResponseAgentPluginsChat(BaseModel):
    class Config:
        json_schema_extra = {
            "example": {
                "message": 'To politely greet shop employees when entering in French, you can say "Bonjour" which means "Hello" or "Good day" in English. It is a common and polite way to greet people in France.',
                "usage": {
                    "total_tokens": 13395,
                    "prompt_tokens": 13241,
                    "completion_tokens": 154,
                    "total_cost": 0.040339,
                    "successful_requests": 4,
                },
            }
        }


#################################################
## Chat Vectorstore HTTP Response
#################################################
class ResponseVectorstoreChat(BaseModel):
    class Config:
        json_schema_extra = {
            "example": {
                "message": "The context is about projects and their components. The components include forms, resources, submissions, actions, logs, access, revisions, settings, roles, stages, integrations, authentication, and stage versions and deployments. The main purpose of the context is to provide a sandbox for projects to manage and control various aspects of their operations. Additionally, users can introduce their own evaluation context variables to customize the evaluations within the projects.",
                "usage": {
                    "total_tokens": 537,
                    "prompt_tokens": 453,
                    "completion_tokens": 84,
                    "total_cost": 0.0008475,
                    "successful_requests": 1,
                },
            }
        }


#################################################
## Create Vectorstore HTTP Response
#################################################
class ResponseCreateVectorStore(BaseModel):
    class Config:
        json_schema_extra = {
            "example": {
                "message": "Vectorstore Created!",
                "data": {
                    "name": "formio-docs-and-website",
                    "loaders": [
                        {"type": "gitbook", "urls": ["https://help.form.io"]},
                        {"type": "website", "urls": ["https://form.io"]},
                    ],
                },
            }
        }


#################################################
## List Vectorstores HTTP Response
#################################################
class ResponseListPineconeVectorStores(BaseModel):
    class Config:
        json_schema_extra = {
            "example": {"vectorstores": ["pinecone-docs-guide-and-api", "formio"]}
        }



"""Routes to get chat history"""
import json
import traceback

from fastapi import APIRouter, HTTPException, Response, Depends, Request, status
from promptengineers.core.interfaces.controllers import IController
from promptengineers.fastapi.controllers import SettingsController
from promptengineers.models.request import ReqBodySettings
from promptengineers.models.response import (ResponseSettingsList, ResponseSetting,
                                            ResponseCreate, ResponseUpdate)
from promptengineers.mongo.utils import JSONEncoder
from promptengineers.core.utils import logger

router = APIRouter()
TAG = "Chat"

def get_controller(request: Request) -> IController:
	return SettingsController(request=request)

#################################################
# List Chat Histories
#################################################
@router.get(
	"/chat/settings",
	tags=[TAG],
	response_model=ResponseSettingsList
)
async def index(
	page: int = 1,
	limit: int = 50,
	controller: IController = Depends(get_controller),
):
	"""List settings"""
	try:
		result = await controller.index(page, limit)
		# Format Response
		data = json.dumps({
			'settings': result,
		}, cls=JSONEncoder)
		return Response(
			content=data,
			media_type='application/json',
			status_code=200
		)
	except HTTPException as err:
		logger.error(err.detail)
		raise
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[controllers.settings.index]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

#################################################
# Create Chat History
#################################################
@router.post(
	"/chat/settings",
	tags=[TAG],
	response_model=ResponseCreate
)
async def create(
	body: ReqBodySettings,
	controller: IController = Depends(get_controller)
):
	"""Creates settings"""
	try:
		result = await controller.create(body)
		# Format Response
		data = json.dumps({
			**result
		})
		return Response(
			content=data,
			media_type='application/json',
			status_code=200
		)
	except HTTPException as err:
		logger.error(err.detail)
		raise
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[controllers.settings.create]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

#################################################
# Show Chat History
#################################################
@router.get(
	"/chat/settings/{setting_id}",
	tags=[TAG],
	response_model=ResponseSetting,
)
async def show(
    setting_id: str,
    controller: IController = Depends(get_controller),
):
	"""Retrieve settings"""
	try:
		result = await controller.show(setting_id)

		# Format Response
		data = json.dumps({
			**result
		}, cls=JSONEncoder)
		return Response(
			content=data,
			media_type='application/json',
			status_code=200
		)
	except HTTPException as err:
		logger.error("%s", err.detail)
		raise
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[controllers.settings.show]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

#################################################
# Update Chat History
#################################################
@router.put(
	"/chat/settings/{setting_id}",
	tags=[TAG],
	response_model=ResponseUpdate,
)
async def update(
	setting_id: str,
	body: ReqBodySettings,
	controller: IController = Depends(get_controller),
):
	"""Update settings"""
	try:
		await controller.update(setting_id, body)
		data = json.dumps({
			'message': f'Setting [{setting_id}] updated successfully.'
		})
		# Format Response
		return Response(status_code=200, content=data)
	except HTTPException as err:
		logger.error("%s", err.detail)
		raise
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[controllers.settings.update]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

#################################################
# Delete Chat History
#################################################
@router.delete(
	"/chat/settings/{setting_id}",
	tags=[TAG],
	status_code=status.HTTP_204_NO_CONTENT,
)
async def delete(
	setting_id: str,
	controller: IController = Depends(get_controller),
):
	"""Delete settings"""
	try:
		await controller.delete(setting_id)
		# Format Response
		return Response(status_code=204)
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[controllers.settings.delete]: %s\n%s", err, tb)
		raise HTTPException(status_code=404, detail=str(err)) from err


import json

from cryptography.fernet import Fernet
from bson import ObjectId

from promptengineers.core.config import APP_SECRET

def encrypt(data: str) -> str:
	return Fernet(APP_SECRET).encrypt(data.encode()).decode()

def decrypt(data: str) -> str:
	return Fernet(APP_SECRET).decrypt(data.encode()).decode()

class JSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, ObjectId):
            return str(o)
        return json.JSONEncoder.default(self, o)


"""Response model for a chat stream message."""

RESPONSE_STREAM_CHAT = """data: {"sender":"assistant","message":"The","type":"stream"}

data: {"sender":"assistant","message":" Arizona","type":"stream"}

data: {"sender":"assistant","message":" Diamond","type":"stream"}

data: {"sender":"assistant","message":"backs","type":"stream"}

data: {"sender":"assistant","message":" had","type":"stream"}

data: {"sender":"assistant","message":" a","type":"stream"}

data: {"sender":"assistant","message":" few","type":"stream"}

data: {"sender":"assistant","message":" standout","type":"stream"}

data: {"sender":"assistant","message":" pitchers","type":"stream"}

data: {"sender":"assistant","message":" during","type":"stream"}

data: {"sender":"assistant","message":" the","type":"stream"}

data: {"sender":"assistant","message":" ","type":"stream"}

data: {"sender":"assistant","message":"200","type":"stream"}

data: {"sender":"assistant","message":"1","type":"stream"}

data: {"sender":"assistant","message":" World","type":"stream"}

data: {"sender":"assistant","message":" Series","type":"stream"}

data: {"sender":"assistant","message":".","type":"stream"}

data: {"sender":"assistant","message":" Randy","type":"stream"}

data: {"sender":"assistant","message":" Johnson","type":"stream"}

data: {"sender":"assistant","message":" and","type":"stream"}

data: {"sender":"assistant","message":" Curt","type":"stream"}

data: {"sender":"assistant","message":" Sch","type":"stream"}

data: {"sender":"assistant","message":"illing","type":"stream"}

data: {"sender":"assistant","message":" were","type":"stream"}

data: {"sender":"assistant","message":" the","type":"stream"}

data: {"sender":"assistant","message":" primary","type":"stream"}

data: {"sender":"assistant","message":" pitchers","type":"stream"}

data: {"sender":"assistant","message":" for","type":"stream"}

data: {"sender":"assistant","message":" the","type":"stream"}

data: {"sender":"assistant","message":" team","type":"stream"}

data: {"sender":"assistant","message":".","type":"stream"}

data: {"sender":"assistant","message":"","type":"end"}"""

#############################################################################
## Langchain Chat Stream Response
#############################################################################
RESPONSE_STREAM_AGENT_CHAT = """data: {"sender":"assistant","message":"$","type":"stream"}

data: {"sender":"assistant","message":"200","type":"stream"}

data: {"sender":"assistant","message":"0","type":"stream"}

data: {"sender":"assistant","message":" compounded","type":"stream"}

data: {"sender":"assistant","message":" at","type":"stream"}

data: {"sender":"assistant","message":" ","type":"stream"}

data: {"sender":"assistant","message":"5","type":"stream"}

data: {"sender":"assistant","message":" percent","type":"stream"}

data: {"sender":"assistant","message":" for","type":"stream"}

data: {"sender":"assistant","message":" ","type":"stream"}

data: {"sender":"assistant","message":"10","type":"stream"}

data: {"sender":"assistant","message":" years","type":"stream"}

data: {"sender":"assistant","message":" will","type":"stream"}

data: {"sender":"assistant","message":" be","type":"stream"}

data: {"sender":"assistant","message":" approximately","type":"stream"}

data: {"sender":"assistant","message":" $","type":"stream"}

data: {"sender":"assistant","message":"325","type":"stream"}

data: {"sender":"assistant","message":"7","type":"stream"}

data: {"sender":"assistant","message":".","type":"stream"}

data: {"sender":"assistant","message":"79","type":"stream"}

data: {"sender":"assistant","message":".","type":"stream"}

data: {"sender":"assistant","message":"","type":"end"}"""

#############################################################################
## Agent Plugins
#############################################################################
RESPONSE_STREAM_AGENT_PLUGINS_CHAT = """data: {"sender":"assistant","message":"When","type":"stream"}

data: {"sender":"assistant","message":" entering","type":"stream"}

data: {"sender":"assistant","message":" a","type":"stream"}

data: {"sender":"assistant","message":" shop","type":"stream"}

data: {"sender":"assistant","message":" in","type":"stream"}

data: {"sender":"assistant","message":" French","type":"stream"}

data: {"sender":"assistant","message":",","type":"stream"}

data: {"sender":"assistant","message":" it","type":"stream"}

data: {"sender":"assistant","message":" is","type":"stream"}

data: {"sender":"assistant","message":" polite","type":"stream"}

data: {"sender":"assistant","message":" to","type":"stream"}

data: {"sender":"assistant","message":" greet","type":"stream"}

data: {"sender":"assistant","message":" the","type":"stream"}

data: {"sender":"assistant","message":" shop","type":"stream"}

data: {"sender":"assistant","message":" employees","type":"stream"}

data: {"sender":"assistant","message":" with","type":"stream"}

data: {"sender":"assistant","message":" a","type":"stream"}

data: {"sender":"assistant","message":" simple","type":"stream"}

data: {"sender":"assistant","message":" \"","type":"stream"}

data: {"sender":"assistant","message":"Bonjour","type":"stream"}

data: {"sender":"assistant","message":"\"","type":"stream"}

data: {"sender":"assistant","message":" (","type":"stream"}

data: {"sender":"assistant","message":"which","type":"stream"}

data: {"sender":"assistant","message":" means","type":"stream"}

data: {"sender":"assistant","message":" \"","type":"stream"}

data: {"sender":"assistant","message":"Hello","type":"stream"}

data: {"sender":"assistant","message":"\"","type":"stream"}

data: {"sender":"assistant","message":" or","type":"stream"}

data: {"sender":"assistant","message":" \"","type":"stream"}

data: {"sender":"assistant","message":"Good","type":"stream"}

data: {"sender":"assistant","message":" day","type":"stream"}

data: {"sender":"assistant","message":"\").","type":"stream"}

data: {"sender":"assistant","message":" Additionally","type":"stream"}

data: {"sender":"assistant","message":",","type":"stream"}

data: {"sender":"assistant","message":" you","type":"stream"}

data: {"sender":"assistant","message":" can","type":"stream"}

data: {"sender":"assistant","message":" use","type":"stream"}

data: {"sender":"assistant","message":" \"","type":"stream"}

data: {"sender":"assistant","message":"Bon","type":"stream"}

data: {"sender":"assistant","message":"so","type":"stream"}

data: {"sender":"assistant","message":"ir","type":"stream"}

data: {"sender":"assistant","message":"\"","type":"stream"}

data: {"sender":"assistant","message":" (","type":"stream"}

data: {"sender":"assistant","message":"which","type":"stream"}

data: {"sender":"assistant","message":" means","type":"stream"}

data: {"sender":"assistant","message":" \"","type":"stream"}

data: {"sender":"assistant","message":"Good","type":"stream"}

data: {"sender":"assistant","message":" evening","type":"stream"}

data: {"sender":"assistant","message":"\")","type":"stream"}

data: {"sender":"assistant","message":" if","type":"stream"}

data: {"sender":"assistant","message":" you","type":"stream"}

data: {"sender":"assistant","message":" enter","type":"stream"}

data: {"sender":"assistant","message":" the","type":"stream"}

data: {"sender":"assistant","message":" shop","type":"stream"}

data: {"sender":"assistant","message":" in","type":"stream"}

data: {"sender":"assistant","message":" the","type":"stream"}

data: {"sender":"assistant","message":" evening","type":"stream"}

data: {"sender":"assistant","message":".","type":"stream"}

data: {"sender":"assistant","message":" These","type":"stream"}

data: {"sender":"assistant","message":" greetings","type":"stream"}

data: {"sender":"assistant","message":" are","type":"stream"}

data: {"sender":"assistant","message":" considered","type":"stream"}

data: {"sender":"assistant","message":" polite","type":"stream"}

data: {"sender":"assistant","message":" and","type":"stream"}

data: {"sender":"assistant","message":" respectful","type":"stream"}

data: {"sender":"assistant","message":" in","type":"stream"}

data: {"sender":"assistant","message":" French","type":"stream"}

data: {"sender":"assistant","message":" culture","type":"stream"}

data: {"sender":"assistant","message":".","type":"stream"}

data: {"sender":"assistant","message":"","type":"end"}"""

#############################################################################
## Langchain Vectorstore Chat Stream
#############################################################################
RESPONSE_STREAM_VECTORSTORE_CHAT = """data: {"sender":"assistant","message":"The","type":"stream"}

data: {"sender":"assistant","message":" context","type":"stream"}

data: {"sender":"assistant","message":" is","type":"stream"}

data: {"sender":"assistant","message":" about","type":"stream"}

data: {"sender":"assistant","message":" projects","type":"stream"}

data: {"sender":"assistant","message":" and","type":"stream"}

data: {"sender":"assistant","message":" their","type":"stream"}

data: {"sender":"assistant","message":" components","type":"stream"}

data: {"sender":"assistant","message":",","type":"stream"}

data: {"sender":"assistant","message":" such","type":"stream"}

data: {"sender":"assistant","message":" as","type":"stream"}

data: {"sender":"assistant","message":" forms","type":"stream"}

data: {"sender":"assistant","message":",","type":"stream"}

data: {"sender":"assistant","message":" resources","type":"stream"}

data: {"sender":"assistant","message":",","type":"stream"}

data: {"sender":"assistant","message":" submissions","type":"stream"}

data: {"sender":"assistant","message":",","type":"stream"}

data: {"sender":"assistant","message":" actions","type":"stream"}

data: {"sender":"assistant","message":",","type":"stream"}

data: {"sender":"assistant","message":" logs","type":"stream"}

data: {"sender":"assistant","message":",","type":"stream"}

data: {"sender":"assistant","message":" access","type":"stream"}

data: {"sender":"assistant","message":",","type":"stream"}

data: {"sender":"assistant","message":" revisions","type":"stream"}

data: {"sender":"assistant","message":",","type":"stream"}

data: {"sender":"assistant","message":" settings","type":"stream"}

data: {"sender":"assistant","message":",","type":"stream"}

data: {"sender":"assistant","message":" roles","type":"stream"}

data: {"sender":"assistant","message":",","type":"stream"}

data: {"sender":"assistant","message":" stages","type":"stream"}

data: {"sender":"assistant","message":",","type":"stream"}

data: {"sender":"assistant","message":" integr","type":"stream"}

data: {"sender":"assistant","message":"ations","type":"stream"}

data: {"sender":"assistant","message":",","type":"stream"}

data: {"sender":"assistant","message":" authentication","type":"stream"}

data: {"sender":"assistant","message":",","type":"stream"}

data: {"sender":"assistant","message":" and","type":"stream"}

data: {"sender":"assistant","message":" stage","type":"stream"}

data: {"sender":"assistant","message":" versions","type":"stream"}

data: {"sender":"assistant","message":" and","type":"stream"}

data: {"sender":"assistant","message":" deployments","type":"stream"}

data: {"sender":"assistant","message":".","type":"stream"}

data: {"sender":"assistant","message":" It","type":"stream"}

data: {"sender":"assistant","message":" also","type":"stream"}

data: {"sender":"assistant","message":" mentions","type":"stream"}

data: {"sender":"assistant","message":" the","type":"stream"}

data: {"sender":"assistant","message":" ability","type":"stream"}

data: {"sender":"assistant","message":" to","type":"stream"}

data: {"sender":"assistant","message":" introduce","type":"stream"}

data: {"sender":"assistant","message":" custom","type":"stream"}

data: {"sender":"assistant","message":" evaluation","type":"stream"}

data: {"sender":"assistant","message":" context","type":"stream"}

data: {"sender":"assistant","message":" variables","type":"stream"}

data: {"sender":"assistant","message":" for","type":"stream"}

data: {"sender":"assistant","message":" validations","type":"stream"}

data: {"sender":"assistant","message":" and","type":"stream"}

data: {"sender":"assistant","message":" other","type":"stream"}

data: {"sender":"assistant","message":" purposes","type":"stream"}

data: {"sender":"assistant","message":".","type":"stream"}

data: {"sender":"assistant","message":"","type":"end"}"""


from pydantic import BaseModel, Field

class SystemMessage(BaseModel):
	"""A message from the system."""
	role: str = Field(example='system')
	content: str = Field(example='You are a helpful assistant.')

class UserMessage(BaseModel):
	"""A message from the user."""
	role: str = Field(example='user')
	content: str = Field(example='Who won the 2001 world series?')

class AssistantMessage(BaseModel):
	"""A message from the assistant."""
	role: str = Field(example='assistant')
	content: str = Field(example='The arizona diamondbacks won the 2001 world series.')


from abc import ABC, abstractmethod
from typing import Any, Dict

class IController(ABC):
    @abstractmethod
    async def index(self, page: int, limit: int) -> Any:
        pass

    @abstractmethod
    async def create(self, body: Dict) -> Any:
        pass

    @abstractmethod
    async def show(self, id: str) -> Any:
        pass

    @abstractmethod
    async def update(self, id: str, body: Dict) -> Any:
        pass

    @abstractmethod
    async def delete(self, id: str) -> Any:
        pass
